/*! \file
 *  \brief Заголовочный файл со структурами и функциями для управления вычислениями
 */

#ifndef _FFT
#define _FFT

#include "DMAInit.h"
#include "elcore50.h"
#include "types_converter.h"

//! Тип данных преобразования
typedef enum {
    FLOAT = 1,      ///< float32
    FRACTIONAL = 2, ///< fractional16
    HFLOAT = 3      ///< float16
} DataTypes;

//! Максимальный размер сигнала, помещающегося во внутреннюю память
#define MAX_FFT_BUF_SIZE 512*2*16*4
//! Размер буфера поворотов для максимального сигнала
#define MAX_FFT_W_BUF_SIZE (3*512*4)/4 
//! Максимальный размер буфера во внутренней памяти
#define MAX_FFT_PLACE_SIZE 16384*2*4

#pragma pack(push, 1)

//! Массивы данных, выгружаемых во внутреннюю память с dma
struct tdma_bufs {
    char x_d0[MAX_FFT_BUF_SIZE] __attribute__((aligned(MAX_FFT_BUF_SIZE*4))); ///< первый входной блок
    char x_r0[MAX_FFT_BUF_SIZE] __attribute__((aligned(MAX_FFT_BUF_SIZE*4))); ///< первый выходной блок
    char x_d1[MAX_FFT_BUF_SIZE] __attribute__((aligned(MAX_FFT_BUF_SIZE*4))); ///< второй входной блок
    char x_r1[MAX_FFT_BUF_SIZE] __attribute__((aligned(MAX_FFT_BUF_SIZE*4))); ///< второй выходной блок
    char t_d0[MAX_FFT_BUF_SIZE] __attribute__((aligned(MAX_FFT_BUF_SIZE*4)));    ///< первый блок поворотов
    char t_d1[MAX_FFT_BUF_SIZE] __attribute__((aligned(MAX_FFT_BUF_SIZE*4)));    ///< второй блок поворотов
    char w_d[MAX_FFT_W_BUF_SIZE] __attribute__((aligned(MAX_FFT_BUF_SIZE*4)));   ///< блок поворотов столбцов/строк
};

//! Входной и выходной массивы во внутренней памяти
struct tinplace_bufs{
    char src[MAX_FFT_PLACE_SIZE] __attribute__((aligned(MAX_FFT_PLACE_SIZE))); ///< входной сигнал
    char dst[MAX_FFT_PLACE_SIZE] __attribute__((aligned(MAX_FFT_PLACE_SIZE))); ///< результат преобразования
};

#pragma pack(pop)

//! Структуры для вычислений во внутренней памяти
union tint_mem_bufs {
    struct tdma_bufs dma_bufs;          ///< для вычислений с dma
    struct tinplace_bufs inplace_bufs;  ///< для вычислений без dma
};

//! Размерность пространства
enum Dimension {
    D1 = 1,     ///< одномерное 
    D2 = 2      ///< двумерное
};

//! Объект преобразования
typedef struct {
    int dim;            ///< размерность пространства
    int size;           ///< размер преобразования
    int height;         ///< количество строк матрицы при переходе к 2D представлению
    int length;         ///< количество столбцов матрицы при переходе к 2D представлению
    void *w_cols;       ///< вектор поворотов столбцов
    void *w_rows;       ///< вектор поворотов строк
    void *t;            ///< вектор поворотов промежуточного этапа
    void *src;          ///< входной сигнал
    void *dst;          ///< результат преобразования
    int dir;            ///< флаг типа преобразования: 1 - прямое, 0 - обратное
    int big_size;       ///< флаг преобразования большого размера: > 16K для типа float32, > 32K для типов fractional16 и float16
    DataTypes fft_type; ///< тип данных преобразования
    int element_size;   ///< размер типа в байтах
    int int_mem;        ///< флаг типа вычислений: 1 - во внутренней памяти, 0 - во внешней
    int ticks;      ///< количество тактов
    int instrs;      ///< количество инструкций
} fft_t;

//! Цепочки dma
typedef struct 
{
    VDMAChain ch_in;        ///< цепочка входных блоков
    VDMAChain ch_out;       ///< цепочка выходных блоков
    VDMAChain ch_twiddles;  ///< цепочка блоков поворотов
} fft_dma_chains;

//! Тип вычислений
enum CALC_MEM {
    EXT_EXT = 0,    ///< во внешней памяти
    EXT_INT = 1     ///< во внутренней памяти
};
//! Тип преобразования
typedef enum {
    IFFT = 0,   ///< обратное
    FFT = 1     ///< прямое
} TransformType;

/*!
 *  \fn int fft_plan_1d(fft_t* fft_obj, int size, DataTypes fft_type)
 *  \brief Создание одномерного объекта преобразования
 *  \param[in] fft_obj входные данные типа fft_t
 *  \param[in] size размер преобразования
 *  \param[in] fft_type тип данных преобразования
 *  \return
 *      - \b 0 нет ошибок
 *      - \b -1 выполнение завершилось с ошибкой
 */
int fft_plan_1d(fft_t* fft_obj, int size, DataTypes fft_type);
/*!
 *  \fn int fft_plan_2d(fft_t *fft_obj, const int height, const int length, DataTypes fft_type)
 *  \brief Создание двумерного объекта преобразования
 *  \param[in] fft_obj входные данные типа fft_t
 *  \param[in] height высота матрицы сигнала
 *  \param[in] length высота матрицы сигнала
 *  \param[in] fft_type тип данных преобразования
 *  \return
 *      - \b 0 нет ошибок
 *      - \b -1 выполнение завершилось с ошибкой
 */
int fft_plan_2d(fft_t *fft_obj, const int height, const int length, DataTypes fft_type);
/*!
 *  \fn void fft_execute(fft_t* fft_obj)
 *  \brief Управление вычислениями
 *  \param[in] fft_obj входные данные типа fft_t
 */
void fft_execute(fft_t* fft_obj);
/*!
 *  \fn void fft_free_sources(fft_t* fft_obj)
 *  \brief Очищение памяти после вычислений
 *  \param[in] fft_obj входные данные типа fft_t
 */
void fft_free_sources(fft_t* fft_obj);
/*!
 *  \fn int create_vdma_fft_chain(const fft_t *fft_obj, fft_dma_chains *chain)
 *  \brief Создание dma цепочки
 *  \param[in] fft_obj входные данные типа fft_t
 *  \param[out] chain выходные данные типа fft_dma_chains
 *  \return
 *      - \b 0 нет ошибок
 *      - \b -1 выполнение завершилось с ошибкой
 */
int create_vdma_fft_chain(const fft_t *fft_obj, fft_dma_chains *chain);

/*!
 *  \fn void fft_float_cols_inter(void *dst, void *src, int height, const void *twiddles, int length, void *temp, int stage)
 *  \brief Вычисление БПФ по столбцам/строкам для типа float
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные 
 *  \param[in] height количество строк входной матрицы сигнала
 *  \param[in] twiddles вектор поворотов столбцов/строк
 *  \param[in] length количество столбцов входной матрицы сигнала
 *  \param[in] temp копия выходной матрицы
 *  \param[in] stage номер этапа преобразования
 */
extern void fft_float_cols_inter(void *dst, void *src, int height, const void *twiddles, int length, 
                                 void *temp, int stage);
/*!
 *  \fn void fft_float_cols_inter_dma_tiles(void *dst, void *src, int height, void *twiddles, int length, void *temp, int stage)
 *  \brief Вычисление БПФ по столбцам/строкам для типа float с использованием dma
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные 
 *  \param[in] height количество строк входной матрицы сигнала
 *  \param[in] twiddles вектор поворотов столбцов/строк
 *  \param[in] length количество столбцов входной матрицы сигнала
 *  \param[in] temp копия выходной матрицы
 *  \param[in] stage номер этапа преобразования
 */
extern void fft_float_cols_inter_dma_tiles(void *dst, void *src, int height, void *twiddles, 
                                           int length, void *temp, int stage);
/*!
 *  \fn void fft_fract_cols_inv(void *dst, void *src, int height, void *twiddles, int length, void *temp, int stage)
 *  \brief Вычисление обратного БПФ по столбцам/строкам для типа fractional16
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные 
 *  \param[in] height количество строк входной матрицы сигнала
 *  \param[in] twiddles вектор поворотов столбцов/строк
 *  \param[in] length количество столбцов входной матрицы сигнала
 *  \param[in] temp копия выходной матрицы
 *  \param[in] stage номер этапа преобразования
 */
extern void fft_fract_cols_inv(void *dst, void *src, int height, void *twiddles, int length, 
                           void *temp, int stage);
/*!
 *  \fn void fft_fract_cols(void *dst, void *src, int height, void *twiddles, int length, void *temp, int stage)
 *  \brief Вычисление прямого БПФ по столбцам/строкам для типа fractional16
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные 
 *  \param[in] height количество строк входной матрицы сигнала
 *  \param[in] twiddles вектор поворотов столбцов/строк
 *  \param[in] length количество столбцов входной матрицы сигнала
 *  \param[in] temp копия выходной матрицы
 *  \param[in] stage номер этапа преобразования
 */
extern void fft_fract_cols(void *dst, void *src, int height, void *twiddles, int length, 
                           void *temp, int stage);
/*!
 *  \fn void fft_hfloat_cols(void *dst, void *src, int height, void *twiddles, int length, void *temp, int stage)
 *  \brief Вычисление БПФ по столбцам/строкам для типа float16
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные 
 *  \param[in] height количество строк входной матрицы сигнала
 *  \param[in] twiddles вектор поворотов столбцов/строк
 *  \param[in] length количество столбцов входной матрицы сигнала
 *  \param[in] temp копия выходной матрицы
 *  \param[in] stage номер этапа преобразования
 */
extern void fft_hfloat_cols(void *dst, void *src, int height, void *twiddles, int length, 
                            void *temp, int stage);

/*!
 *  \fn void fft_twiddles_cols_gen(void *w, int size)
 *  \brief Вычисление поворотов столбцов/строк для типа float
 *  \param[out] w выходные данные
 *  \param[in] size размер входных данных
 */
extern void fft_twiddles_cols_gen(void *w, int size);
/*!
 *  \fn void fft_twiddles_t_gen_inter(void *t, int size, int height, int length)
 *  \brief Вычисление поворотов для типа float
 *  \param[out] t выходные данные
 *  \param[in] size размер входных данных
 *  \param[in] height количество строк входной матрицы
 *  \param[in] length количество столбцов входной матрицы
 */
extern void fft_twiddles_t_gen_inter(void *t, int size, int height, int length);

/*!
 *  \fn void fft_fract_twiddles_cols_gen(void *w, int size)
 *  \brief Вычисление поворотов столбцов/строк для типа fractional16
 *  \param[out] w выходные данные
 *  \param[in] size размер входных данных
 */
extern void fft_fract_twiddles_cols_gen(void *w, int size);
/*!
 *  \fn void fft_fract_twiddles_t_gen(void *t, int size, int height, int length)
 *  \brief Вычисление поворотов для типа fractional16
 *  \param[out] t выходные данные
 *  \param[in] size размер входных данных
 *  \param[in] height количество строк входной матрицы
 *  \param[in] length количество столбцов входной матрицы
 */
extern void fft_fract_twiddles_t_gen(void *t, int size, int height, int length);

/*!
 *  \fn void fft_hfloat_twiddles_cols_gen(void  *w, int size)
 *  \brief Вычисление поворотов столбцов/строк для типа float16
 *  \param[out] w выходные данные
 *  \param[in] size размер входных данных
 */
extern void fft_hfloat_twiddles_cols_gen(void  *w, int size);
/*!
 *  \fn void fft_hfloat_twiddles_t_gen(void *t, int size, int height, int length)
 *  \brief Вычисление поворотов для типа float16
 *  \param[out] t выходные данные
 *  \param[in] size размер входных данных
 *  \param[in] height количество строк входной матрицы
 *  \param[in] length количество столбцов входной матрицы
 */
extern void fft_hfloat_twiddles_t_gen(void *t, int size, int height, int length);

/*!
 *  \fn void fft_mult_and_transpose_inter(float *dst, float *src, int src_height, int src_length, float *T)
 *  \brief Умножение на матрицу поворотов и транспонирование матрицы типа float 
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] src_height количество строк входной матрицы
 *  \param[in] src_length количество столбцов входной матрицы
 *  \param[in] T матрица поворотов
 */
extern void fft_mult_and_transpose_inter(float *dst, float *src, int src_height, int src_length, float *T);
/*!
 *  \fn void fft_mult_and_transpose(void  *dst, void  *src, int src_height, int src_length, void *T)
 *  \brief Умножение на матрицу поворотов и транспонирование матрицы типа float
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] src_height количество строк входной матрицы
 *  \param[in] src_length количество столбцов входной матрицы
 *  \param[in] T матрица поворотов
 */
extern void fft_mult_and_transpose(void  *dst, void  *src, int src_height, int src_length, void *T);
/*!
 *  \fn void fft_mult_and_transpose_fract(void *dst, void  *src, int src_height, int src_length, void *T)
 *  \brief Умножение на матрицу поворотов и транспонирование матрицы типа fractional16
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] src_height количество строк входной матрицы
 *  \param[in] src_length количество столбцов входной матрицы
 *  \param[in] T матрица поворотов
 */
extern void fft_mult_and_transpose_fract(void *dst, void  *src, int src_height, int src_length, void *T);
/*!
 *  \fn void fft_mult_and_transpose_hfloat(void  *dst, void  *src, int src_height, int src_length, void *T)
 *  \brief Умножение на матрицу поворотов и транспонирование матрицы типа float16
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] src_height количество строк входной матрицы
 *  \param[in] src_length количество столбцов входной матрицы
 *  \param[in] T матрица поворотов
 */
extern void fft_mult_and_transpose_hfloat(void  *dst, void  *src, int src_height, int src_length, void *T);

/*!
 *  \fn void fft_transpose_inter(float *dst, float *src, int src_height, int src_length)
 *  \brief Транспонирование матрицы типа float 
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] src_height количество строк входной матрицы
 *  \param[in] src_length количество столбцов входной матрицы
 */
extern void fft_transpose_inter(float *dst, float *src, int src_height, int src_length);
/*!
 *  \fn void fft_transpose_fract(float *dst, float *src, int src_height, int src_length)
 *  \brief Транспонирование матрицы типа fractional16 
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] src_height количество строк входной матрицы
 *  \param[in] src_length количество столбцов входной матрицы
 */
extern void fft_transpose_fract(float *dst, float *src, int src_height, int src_length);
/*!
 *  \fn void fft_transpose_hfloat(float *dst, float *src, int src_height, int src_length)
 *  \brief Транспонирование матрицы типа float16 
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] src_height количество строк входной матрицы
 *  \param[in] src_length количество столбцов входной матрицы
 */
extern void fft_transpose_hfloat(float *dst, float *src, int src_height, int src_length);

#endif
